#!/usr/bin/env python3
"""Input source utility."""


class InputSource(object):
    """Object definition for InputSource."""

    def __iter__(self):
        """Iterate over log events."""
        pass

    def fast_forward(self, dt):
        """Fast forward over log events."""
        pass
